#!/usr/local/bin/python
# @(#) $Id$ (LBL)
"""logcount-often - mange periodic logcount runs"""

import argparse
import errno
import fcntl
import json
import os
import subprocess
import sys
import time

OPTS = None
PROG = '?'

DEFAULT_DATAFILE = '/var/db/logcount/logcount.json'

class database:
    """logcount database class"""
    def __init__(self, fn, readonly=False):
        """Instantiate the class"""
        self._f = None
        self._fn = fn
        self._readonly = readonly
        self.state = None

    def open(self):
        """Open and read the database,
           lock if not read/only,
           return an optional error message"""
        if self._readonly:
            mode = 'r'
        else:
            mode = 'r+'
        newdatabase = False
        try:
            # pylint: disable=consider-using-with
            self._f = open(self._fn, mode, encoding='utf-8')
            # pylint: enable=consider-using-with
        except (IOError, OSError) as e:
            if self._readonly or e.errno != errno.ENOENT:
                errmsg = f'{self._fn}: {e.strerror}'
                return errmsg
            # Database doesn't exist; create a new one
            try:
                # pylint: disable=consider-using-with
                self._f = open(self._fn, 'w+', encoding='utf-8')
                # pylint: enable=consider-using-with
            except (IOError, OSError) as e2:
                errmsg = f'{self._fn}: {e2.strerror}'
                return errmsg
            newdatabase = True

        if not self._readonly:
            try:
                fcntl.flock(self._f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError as e:
                if e.errno == errno.EAGAIN:
                    detail = 'Locked by another instance'
                else:
                    detail = str(e.strerror)
                errmsg = f'{self._fn}: {detail}'
                return errmsg

        # Read the database
        try:
            self.state = json.load(self._f)
        except ValueError:
            if not newdatabase:
                errmsg = 'resetting corrupt database'
            else:
                errmsg = 'created new database'
            print(f'{self._fn}: {errmsg}', file=sys.stderr)
            self.state = {}

        for key in ('inode', 'offset'):
            val = self.state.get(key)
            if not isinstance(val, int):
                print(f'{PROG}: reset "{self._fn}" from bogus value ({val})',
                    file=sys.stderr)
                self.state = {}

        return None

    def close(self):
        """Update the database if not readonly and close
           return an optional error message"""
        if not self._readonly:
            self._f.seek(0)
            self._f.truncate()
            try:
                json.dump(self.state, self._f)
            except IOError as e:
                return str(e)
            self._f.close()
            self._f = None
        return None

def process():
    """Process the syslog file, optionally write a result file and
       update the database,
       return an optional error message"""

    # Open/read the database
    db = database(OPTS.datafile, readonly=OPTS.dryrun)
    errmsg = db.open()
    if errmsg:
        return errmsg

    cmds = ['logcount']
    inode = db.state.get('inode')
    offset = db.state.get('offset')
    if inode is not None:
        cmds.extend(['-I', str(inode)])
    # Use an offset of zero to bootstrap
    if offset is None:
        offset = 0
    cmds.extend(['-O', str(offset)])
    cmds.append(OPTS.syslog)
    if OPTS.debug:
        print(f"+ {' '.join(cmds)}", file=sys.stderr)
    ret, lines, errlines = run(cmds)
    if ret != 0:
        return '\n'.join(errlines)

    if OPTS.debug and errlines:
        for errline in errlines:
            print(f'# {errline}')

    # Remove and parse informational comments
    db.state['inode'] = None
    db.state['offset'] = None
    while lines and lines[0].startswith('#'):
        line = lines.pop(0).lstrip('#').strip()
        tup = line.split()
        if len(tup) != 2:
            print(f'{PROG}: bad logcount comment: {line}',
                file=sys.stderr)
            continue
        if tup[0] == 'inode':
            db.state['inode'] = int(tup[1])
            continue
        if tup[0] == 'offset':
            db.state['offset'] = int(tup[1])
            continue
        print(f'{PROG}: unexpected logcount comment: {line}',
            file=sys.stderr)

    tup = []
    for key in ('inode', 'offset'):
        if db.state.get(key) is None:
            tup.append(f"didn't find {key}")
    if tup:
        errmsg = ', '.join(tup)
        return errmsg

    # Update the database
    errmsg = db.close()
    if errmsg:
        return errmsg

    # Create a new result file if there were any
    if not OPTS.dryrun and lines:
        lines.sort()
        sn = f'{int(time.time())}.txt'
        fn = os.path.join(OPTS.resultdir, sn)
        try:
            with open(fn, 'w', encoding='utf-8') as f:
                for line in lines:
                    print(line, file=f)
        except (IOError, OSError) as e:
            errmsg = f'{fn}: {e.strerror}'
            return errmsg

    return None

def run(cmd):
    """Run a command, return the status, a list list of output lines
       and a list of error lines"""
    try:
        with subprocess.Popen(cmd, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, encoding='utf8') as p:
            tup = p.communicate()
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise e
        return 1, None, [f'{PROG}: {cmd[0]}: {e.strerror}']

    # Split lines into a list and remove empty lines
    lines = [_f for _f in tup[0].splitlines() if _f]
    errlines = [_f for _f in tup[1].splitlines() if _f]
    return p.returncode, lines, errlines

def main(argv=None):
    """Parse options, output config"""
    global OPTS
    global PROG

    if not argv:
        argv = sys.argv

    PROG = os.path.basename(argv[0])
    usage = '%(prog)s [-dn] [-D DATAFILE] SYSLOG RESULTDIR'
    version = '$Revision$'.lstrip('$').rstrip('$')

    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument('-d', dest='debug', action='count', default=0,
        help='turn on debugging')
    parser.add_argument('-n', dest='dryrun', action='store_true',
        help="don't update/create any files (dry run)")
    parser.add_argument('-D', dest='datafile', action='store',
        default=DEFAULT_DATAFILE,
        help='database file (default: %(default)s)')
    parser.add_argument('syslog', metavar='SYSLOG',
        help='input syslog file')
    parser.add_argument('resultdir', metavar='RESULTDIR',
        help='results directory')

    parser.add_argument('--debugger', action='store_true',
        help='interactive debugging')

    OPTS = parser.parse_args()
    # argparse debugging
    if OPTS.debug > 1:
        for key in dir(OPTS):
            if not key.startswith('_'):
                print(f'# {key}={getattr(OPTS, key)}', file=sys.stderr)

    # Interactive debugging
    if OPTS.debugger:
        # pylint: disable=import-outside-toplevel
        import pdb
        # pylint: enable=import-outside-toplevel
        # pylint: disable=forgotten-debug-statement
        pdb.set_trace()
        # pylint: enable=forgotten-debug-statement

    # Make sure our path will find logcount
    pathstr = 'PATH'
    localbin = '/usr/local/bin'
    if pathstr in os.environ and localbin not in os.environ[pathstr]:
        os.environ[pathstr] = f'{localbin}:{os.environ[pathstr]}'

    errmsg = process()
    if errmsg:
        print(f'{PROG}: {errmsg}', file=sys.stderr)
        return 1

    return 0

if __name__ == "__main__":
    sys.exit(main())
