#!/usr/local/bin/python
# @(#) $Id$ (LBL)
"""logcount-report - generate the syslog recent report"""

import argparse
import datetime
import os
import re
import sys
import time

OPTS = None
PROG = '?'

def doone(fn, timestamps, hosts, startts):
    """Process one recent file, updating the hosts dict"""
    try:
        with open(fn, encoding='utf-8') as f:
            # Keep track of newest and oldest timestamps
            s = os.fstat(f.fileno())
            ts = int(s.st_mtime)
            if ts <= startts:
                if OPTS.debug:
                    print(f'# skipping {fn} with mtime {ts}',
                        file=sys.stderr)
                return None

            timestamps.append(ts)
            timestamps.sort()
            if timestamps[0] == 0:
                timestamps.pop(0)
            else:
                timestamps.pop(1)

            n = 0
            for line in [x.rstrip() for x in f.readlines()]:
                n += 1
                if not line or line.startswith('#'):
                    continue
                tup = line.split()
                if len(tup) != 3:
                    print(f'bad line: {fn}.{n}: {line}',
                        file=sys.stderr)
                    continue

                count = 0
                try:
                    count = int(tup[2])
                except ValueError:
                    print(f'bad count: {fn}.{n}: {line}',
                        file=sys.stderr)
                    continue

                hostname = tup[0]
                addr = tup[1]

                key = ' '.join([hostname, addr])
                oldcount = hosts.get(key, 0)
                hosts[key] = oldcount + count

    except (IOError, OSError) as e:
        errmsg = f'{fn}: {e.strerror}'
        return errmsg

    return None

def parsestart(start):
    """Parse HH:MM and return a unix timestamp and an optional error message"""
    if not start:
        return 0, None

    tup = start.split(':', 1)
    errmsg = 'Bad start timestamp format'
    if len(tup) != 2:
        return None, errmsg
    if not all(x.isdigit() for x in tup):
        return None, errmsg

    # Find a time with the specified hour and minute prior to now
    h, m = [int(x) for x in tup]
    now = int(time.time())
    candidate = now
    for _i in range(2):
        thedate = time.localtime(candidate)
        year = thedate.tm_year
        mon = thedate.tm_mon
        mday = thedate.tm_mday
        dt = datetime.datetime(year, mon, mday, h, m)
        ts = int(time.mktime(dt.timetuple()))
        if ts < now:
            if OPTS.debug:
                print(f'# start ts {ts}', file=sys.stderr)
            return ts, None

        candidate -= 24 * 3600

    return None, 'Failed to find a time'

def prettydate(ts=None, dotz=True):
    """Return formatted date and time"""
    thedate = time.localtime(ts)
    fmts = [
        '%Y-%m-%d %H:%M',
    ]
    if dotz:
        fmts.append('%Z')
    return time.strftime(' '.join(fmts), thedate)

def process(dn, startts):
    """Process files in the recent directory, write the report to stdout.
       return an optional error message"""

    hosts = {}
    timestamps = [0, 0]
    n = 0
    try:
        for sn in os.listdir(dn):
            if not sn.endswith('.txt'):
                continue
            fn = os.path.join(dn, sn)
            errmsg = doone(fn, timestamps, hosts, startts)
            if errmsg:
                return errmsg
            n += 1
    except OSError as e:
        errmsg = f'{dn}: {e.strerror}'
        return errmsg

    # Calculate cron job interval and subtract from first timestamp
    if n > 0:
        delta = timestamps[1] - timestamps[0]
        interval = int(round((1.0 * delta) / (n - 1)))
        if interval < 3600:
            timestamps[0] -= interval

    # Generate report
    fmt1 = '%-40s %-20s %s'
    fmt2 = '%-40s %-20s %d'

    print(f'Updated: {prettydate()}')
    print('')
    print(f'Showing counts from {prettydate(timestamps[0], dotz=False)}'
        f' to {prettydate(timestamps[1])}')
    if OPTS.blurb:
        print('')
        try:
            with open(OPTS.blurb, encoding='utf-8') as f:
                for line in [x.rstrip() for x in f.readlines()]:
                    print(line)
        except (IOError, OSError) as e:
            errmsg = f'{OPTS.blurb}: {e.strerror}'
            return errmsg
    print('')
    head1 = fmt1 % ('Hostname', 'IP', 'Count')
    head2 = re.sub(r'\S', '-', head1)
    print(head1)
    print(head2)
    hostnameaddrs = list(hosts.keys())
    hostnameaddrs.sort()
    for hostnameaddr in hostnameaddrs:
        count = hosts.get(hostnameaddr, 0)
        tup = hostnameaddr.split()
        hostname = tup[0]
        addr = tup[1]
        print(fmt2 % (hostname, addr, count))

    print('')
    print(f'Total Hosts: {len(hostnameaddrs)}')

    return None

def main(argv=None):
    """Parse options, output config"""
    global OPTS
    global PROG

    if not argv:
        argv = sys.argv

    PROG = os.path.basename(argv[0])
    usage = '%(prog)s [-dn] [-b BLURB] [-t HH:MM] RESULTDIR'
    version = '$Revision$'.lstrip('$').rstrip('$')

    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument('-d', dest='debug', action='count', default=0,
        help='turn on debugging')
    parser.add_argument('-n', dest='dryrun', action='store_true',
        help="don't update/create any files (dry run)")
    parser.add_argument('-b', dest='blurb',
        help='optional blurb file')
    parser.add_argument('-t', dest='start',
        help='ignore results before time (example: 23:55)')
    parser.add_argument('resultdir', metavar='RESULTDIR',
        help='results directory')

    parser.add_argument('--debugger', action='store_true',
        help='interactive debugging')

    OPTS = parser.parse_args()
    # argparse debugging
    if OPTS.debug > 1:
        for key in dir(OPTS):
            if not key.startswith('_'):
                print(f'# {key}={getattr(OPTS, key)}', file=sys.stderr)

    # Interactive debugging
    if OPTS.debugger:
        # pylint: disable=import-outside-toplevel
        import pdb
        # pylint: enable=import-outside-toplevel
        # pylint: disable=forgotten-debug-statement
        pdb.set_trace()
        # pylint: enable=forgotten-debug-statement

    startts, errmsg = parsestart(OPTS.start)
    if errmsg:
        print(f'{PROG}: {errmsg}', file=sys.stderr)
        return 1

    errmsg = process(OPTS.resultdir, startts)
    if errmsg:
        print(f'{PROG}: {errmsg}', file=sys.stderr)
        return 1

    return 0

if __name__ == "__main__":
    sys.exit(main())
